#include <wtypes.h>
//#include <msfs.h>
#include <stdio.h>
#include <winsock.h>
#include <winbase.h>
#include <wincon.h>

#define BGPCSRV "billgPC Launcher"
//#define DEBUG_BGPCSVC 1         // deluxe debugging info

void usage()
{
  char* usageText =
    "The " BGPCSRV " is bgpcsvc.exe, an NT service program which can be\n"
    "used to launch billgpc.exe automatically before any users log in\n"
    "when both programs reside in the same directory.\n"
    "The " BGPCSRV " is started from the command line only to install it\n"
    "or remove it as a service.\n"
    "To install the " BGPCSRV " type\n"
    "bgpcsvc create <args>\n"
    "where <args> are any arguments you want passed to billgPC.\n"
    "To remove the " BGPCSRV " type\n"
    "bgpcsvc delete\n"
    "Once the " BGPCSRV " has been installed as a service, if you want to\n"
    "change the arguments it passes to billgPC you must remove it and\n"
    "reinstall it as detailed above.\n";
  printf(usageText);
  exit(-1);
}

int  bgpcsvcMain(int nargs, char *argv[]);
void SendStatus(DWORD stat, DWORD checkpoint);

static FILE* fpOhNo = NULL;

void SendSimpleEvt(char *Server, char *AppName, int Type, short Cat,
                   int ID, const char *str)
{
  HANDLE eHandle = RegisterEventSource(Server, AppName);

  if(eHandle == NULL) {
    if (fpOhNo == NULL) fpOhNo = fopen("bgpcsvc.log", "w");
    fprintf(fpOhNo, "SendSimpleEvt: cannot register, error %d %s\n",
            GetLastError(), "- trying local");

    eHandle = RegisterEventSource("", AppName);
    if(eHandle == NULL) {
    if (fpOhNo == NULL) fpOhNo = fopen("bgpcsvc.log", "w");
      fprintf(fpOhNo, "SendSimpleEvt: cannot register, error %d %s\n",
              GetLastError(), "(bother)");
      return;
    }
  }

  if (ReportEvent(eHandle, Type, Cat, ID, NULL, 1, 0, &str, NULL) == FALSE) {
    if (fpOhNo == NULL) fpOhNo = fopen("bgpcsvc.log", "w");
    fprintf(fpOhNo, "ReportEvent fails, lasterror is %d\n", GetLastError());
  }

  DeregisterEventSource(eHandle);

}


void SendInfo(char *Server, char *AppName,
              short Cat, int ID, char *str)
{
  SendSimpleEvt(Server, AppName, EVENTLOG_INFORMATION_TYPE, Cat, ID, str);
}

void SendWarn(char *Server, char *AppName,
              short Cat, int ID, char *str)
{
  SendSimpleEvt(Server, AppName, EVENTLOG_WARNING_TYPE, Cat, ID, str);
}

void SendError(char *Server, char *AppName,
               short Cat, int ID, char *str)
{
  SendSimpleEvt(Server, AppName, EVENTLOG_ERROR_TYPE, Cat, ID, str);
}

/*
  These simple logging macros place stuff into the NT event log.
*/

#ifdef DEBUG_BGPCSVC
#define mylog(level,str,evtID) \
   SendInfo("", BGPCSRV, level, evtID, str);
#else
#define mylog(level,str,evtID)
#endif

#define mywarn(level,str,evtID)                              \
   SendWarn("", BGPCSRV, level, evtID, str);

#define myerr(str,evtID)                                     \
   SendError("", BGPCSRV, 0, evtID, str)

/* This event is used to notify the service of control events */
static HANDLE HctrlEvent;

/* This flag is used to notify the service to stop */
static int termFlag;

/*
  iStat is used as a "checkpoint" counter.  When we tell the
  Service Manager how we're doing using SendStatus, we increment
  this number just to say that we're doing things.  NT uses this
  to determine if we're stuck, and how long it'll take to get going.
*/
long iStat; 

/* Don't fill up the NT event log if the error keeps happening */
static int loggedErr;   

/*
  Here we launch billgPC.  It should be in the same directory from which
  the service was plucked, and it takes the same arguments.
*/

void launchBillgPC()
{
  PROCESS_INFORMATION myProcInfo = { 0 };
  STARTUPINFO myStartInfo = {
    sizeof(myStartInfo),        // DWORD cb
    NULL,                       // LPSTR lpReserved
#if 0
    "WinSta0\\Default",		// *LPSTR lpDesktop
#else
    NULL,                       // Let it inherit from us!
#endif
    NULL,                       // LPSTR lpTitle; the rest is 0
  };

  /*
    Because of the possibility that billgPC and bgpcsvc's home directory
    name has embedded whitespace, the installation process
    ("bgpcsvc create...") makes sure that the command line will surround
    bgpcsvc's full path name with quotes.  So what we do here is to build
    a command line for billgPC by substituting its filename for that of
    bgpcsvc, preserving the quotes and appending the options, if any, that
    begin with the first forward slash.
  */
  char* cmdLine = GetCommandLine();
  char billgPCCmdLine[200];     // New process's command line
  char* lastBackslash = strchr(cmdLine, '\\');
  while(1) {
    size_t nextSentinelOffset = strcspn(lastBackslash + 1, "\\");
    char sentinel = lastBackslash[nextSentinelOffset + 1];
    if (sentinel != '\\') break;
    lastBackslash += (nextSentinelOffset + 1);
  }
  size_t dirPathLen = lastBackslash - cmdLine + 1;
  memcpy(billgPCCmdLine, cmdLine, dirPathLen);
  strcpy(billgPCCmdLine + dirPathLen, "BILLGPC.EXE\" ");
  char* options = strstr(cmdLine, "/");
  if (options) strcat(billgPCCmdLine, options);

  /*
    billgPC is created as a separate process inheriting a desktop that's
    guaranteed to accept input even if nobody's logged in.  Once it's
    created we restore the desktop we were born with.
  */
  HDESK oldDesktop = GetThreadDesktop(GetCurrentThreadId());
  HDESK newDesktop = OpenInputDesktop(0, TRUE, DESKTOP_CREATEWINDOW);
  SetThreadDesktop(newDesktop);
#if 0
  myerr(billgPCCmdLine, 0);
#endif
  CreateProcess(NULL, billgPCCmdLine, NULL, NULL, 
                TRUE, CREATE_NEW_PROCESS_GROUP | DETACHED_PROCESS, 
                NULL, NULL, 
                &myStartInfo, &myProcInfo);
  SetThreadDesktop(oldDesktop);

  if (myProcInfo.hProcess) {
    /* wait for the desktop window to actually be started */
    WaitForInputIdle(myProcInfo.hProcess, INFINITE);

    CloseHandle(myProcInfo.hProcess);
    CloseHandle(myProcInfo.hThread);
    loggedErr = FALSE;
  }
  else if (!loggedErr) {
    char buf[80];
    wsprintf(buf, "Can't launch desktop window: %d", GetLastError());
    myerr(buf, 12);
    loggedErr = TRUE;
  }        

  return;
}

/*
  Here is the main body of the service.
*/

int bgpcsvcMain(int /* argc */, char ** /* argv */)
{
  iStat = 0;

  termFlag = FALSE;

  /*
    Tell the Service Manager that we've started up.  This is
    extremely important.  If we have a lot of initialization to do,
    then periodically send SERVICE_START_PENDING messages while
    starting up. 
   */
  SendStatus(SERVICE_RUNNING, iStat++);

  mylog(0, "launching billgPC", 7);
  launchBillgPC();
  for (int timeToLive = 20; timeToLive > 0; timeToLive--) {
    /*
      Wait on the control event, with a one second timeout.  if we
      have other jobs, we might create separate threads for them to do
      their work.  They should check termFlag to see when they should
      quit.  
    */
    WaitForSingleObject(HctrlEvent, 1000);
  
    /*
      If the SvcCntrlHandler has been instructed to stop the
      service, then it will set termFlag.
    */
    if (termFlag) break;        /* just drop out of our service */
  }

  /*
    At this point we've been instructed to shut down completely.
    Tell the Service Manager that we are coming down.
  */
  SendStatus(SERVICE_STOP_PENDING, iStat++);

  /* and put an event into the NT Application log */
  mylog(0, "leaving bgpcsvcMain", 8);
  return 0;
}


/*
  The service communicates its status (SERVICE_START_PENDING
  etc.) to the NT service manager  via SetServiceStatus() function 
  call which is passed the sts structure.
*/
SERVICE_STATUS sts;     

/* When the service is opened, it is given a svcHandle */
SERVICE_STATUS_HANDLE svcHandle;

/* 
  Any arguments passed in the "binPath" of the service will be passed
  as arguments to the main() function.  If the service thread should
  use these, it can access the global pointers below.  The "startup
  parameters" specified in the Service Manager, on the other hand, are
  passed as parameters to mainproc() when the Service Manager starts
  it up.
*/
static int gArgc;
static char **gArgv;

/* The server program calls this function to let the Service Manager
  know how it is getting on.  With SERVICE_START_PENDING or
  SERVICE_STOP_PENDING the "checkpoint" is used by the Service Manager
  to see that things are moving along and sts.dwWaitHint tells the
  service manager not to bother it for that many milliseconds.
*/
void SendStatus(DWORD cursts, DWORD checkpoint)
{
  sts.dwCurrentState = cursts;
  sts.dwCheckPoint = checkpoint;
  if ((cursts == SERVICE_START_PENDING) || (cursts == SERVICE_STOP_PENDING)) {
    sts.dwWaitHint = 3000;
  }
  else sts.dwWaitHint = 0;
  if (!SetServiceStatus(svcHandle, &sts)) myerr("Can't send status",
                                                 GetLastError());
}

/*
  svcCntrlHandler

  Each service must contain a service control handler.  This one
  gets the control messages from the Service Manager, and uses
  an event handle and a global flag to notify the service
  whether it should stop.
*/
VOID WINAPI svcCntrlHandler(DWORD cntrlMsg)
{
  if((cntrlMsg == SERVICE_CONTROL_STOP)
     || (cntrlMsg == SERVICE_CONTROL_SHUTDOWN)) {
    SendStatus(SERVICE_STOP_PENDING, 99);
    termFlag = TRUE;
    SetEvent(HctrlEvent);
  }
  else {
    SendStatus(SERVICE_RUNNING, 8);
  }
}

/* 
   The Service Manager starts this routine up as a thread.  It calls
   the service body routine bgpcsvcMain() as a function.  When the
   service exits, flow of control returns to this routine.
*/

VOID WINAPI mainproc(DWORD /* argc */, LPTSTR* /* argv */)
{
  char dbgmsg[40];

  mylog(0,"mainproc called", 3);

  /*
    Set some parms for the status communication structure. This 
    structure is then used by SendStatus(), making it easier for the 
    body of the service to notify the Service Manager of its status.
  */
  sts.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
  sts.dwCurrentState = SERVICE_START_PENDING;
  sts.dwControlsAccepted = SERVICE_ACCEPT_STOP + SERVICE_ACCEPT_SHUTDOWN;
  sts.dwWin32ExitCode = NO_ERROR;
  sts.dwServiceSpecificExitCode = 0;
  sts.dwCheckPoint = 0;
  sts.dwWaitHint = 0;

  /*
    create an event which the SvcCntrlHandler uses to notify 
    the service about stopping.
  */
  HctrlEvent = CreateEvent(NULL, FALSE, FALSE, BGPCSRV);
  if (!HctrlEvent) {
    wsprintf(dbgmsg,"Error %ld.  CreateEvent failed",
             (long)GetLastError());

    /* Put a message into the NT Application event log */
    myerr(dbgmsg, 4);
    return;
  }


  /*
    Tell the Service Manager where to find the service.
  */
  svcHandle = RegisterServiceCtrlHandler(BGPCSRV, svcCntrlHandler);

  if (!svcHandle) {
    wsprintf(dbgmsg,"Error %ld.  RegSvcCtrlHndl failed", GetLastError());

    /* Put a message into the NT Application event log */
    myerr(dbgmsg,5);
    return;
  }

  SendStatus(SERVICE_START_PENDING, 0);

  /*
    Now we execute the body of the service.
  */

  bgpcsvcMain(gArgc, gArgv);  

  CloseHandle(HctrlEvent);

  // and that's it.
  sts.dwCurrentState = SERVICE_STOPPED;
  SetServiceStatus(svcHandle, &sts);
  mylog(0,"mainproc exit",6);
}

SC_HANDLE getSCMHandle(void)
{
  SC_HANDLE scm = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);

  if (scm == NULL) {
    printf("Couldn't open handle to SC Manager (0x%x)\n", GetLastError());
    usage();
  }
  return scm;
}

/*
  Here we create in binPath a command line for the Service Control Manager
  to use in invoking bgpcsvc.  It consists of the program's full path name
  (encased in quotes in case we're situated in a directory whose name
  contains whitespace) followed by the slash options we pass to billgPC.
*/

void createBgpcsvc()
{
  SC_HANDLE scm = getSCMHandle();
  char binPath[_MAX_PATH];
  char rawFilePath[_MAX_PATH];
  static char quotedFilePath[_MAX_PATH];
  char dependencies[] = { 'T', 'c', 'p', 'i', 'p', '\0', '\0' };
  GetModuleFileName(NULL, rawFilePath, sizeof(rawFilePath));
  strcpy(quotedFilePath, "\"");
  strcat(quotedFilePath, rawFilePath);
  strcat(quotedFilePath, "\"");
  strcpy(binPath, quotedFilePath);
  for (int argNo = 2; argNo < gArgc; argNo++) {
    if (*gArgv[argNo] != '/') continue;
    strcat(binPath, " ");
    strcat(binPath, gArgv[argNo]);
  }
#if defined(DEBUG_BGPCSVC)
  printf("binPath = %s\n", binPath);
#endif
  if (!CreateService(scm, BGPCSRV, BGPCSRV, SERVICE_ALL_ACCESS,
                     SERVICE_WIN32_OWN_PROCESS | SERVICE_INTERACTIVE_PROCESS,
                     SERVICE_AUTO_START,
                     SERVICE_ERROR_NORMAL, binPath, NULL, NULL,
                     dependencies, NULL, NULL)) {
    printf("Couldn't create service (0x%x)\n", GetLastError());
    usage();
  }
  else printf(BGPCSRV " successfully created\n");
  CloseServiceHandle(scm);
}

void deleteBgpcsvc()
{
  SC_HANDLE scm = getSCMHandle();
  SC_HANDLE service = OpenService(scm, BGPCSRV, SERVICE_ALL_ACCESS);
  if (service == NULL) {
    printf("Couldn't open " BGPCSRV);
    usage();
  }
  if (DeleteService(service)) printf(BGPCSRV " successfully deleted\n");
  else printf("Couldn't delete " BGPCSRV);
}

SERVICE_TABLE_ENTRY   svcTab[] = {
  { TEXT(BGPCSRV), mainproc      },
  { NULL,            NULL                } 
};

/*
  When the Service Manager calls the service, it enters here.
  Calling StartServiceCtrlDispatcher causes the Service Manager
  to pass control to the mainproc routine of the server, as declared above.
*/
int main(int argc, char *argv[])
{
  char dbmsg[80];

  gArgc = argc;
  gArgv = argv;

  if (argc > 1) {
    if (!lstrcmpi(argv[1], "create")) {
      createBgpcsvc();
      return 0;
    }
    else if (!lstrcmpi(argv[1], "delete")) {
      deleteBgpcsvc();
      return 0;
    }
  }

  mylog(0, "main called", 0);

  if (!StartServiceCtrlDispatcher(svcTab)) {
    /* Put a message in NT App log */
    wsprintf(dbmsg,"Error %ld.  "
             "StartServiceCtrlDispatcher failed", (long)GetLastError());
    myerr(dbmsg,1);
    usage();
  }

  mylog(0, "main exit", 1);

  if (fpOhNo != NULL) fclose(fpOhNo);

  return(0);
}
