/*
   NTBGPS.H - include file for kernel-mode driver BILLGPS

   Free software Copyright (c) 1997 Lew Perin

*/

#ifndef __BILLGPS_H__
#define __BILLGPS_H__

#pragma warning(disable: 4201 4214 4010 4115) // Microsoft is excused...
#include <ndis.h>
#include <ntddndis.h>
#pragma warning(default: 4201 4214 4010 4115) // but we aren't
#include "bgpsdbg.h"
#pragma warning(disable: 4100)  // unreferenced formal parameter
#pragma warning(disable: 4514)  // unreferenced inline function removed

/*
  Now some constants from WINIOCTL.H:
*/

#define FILE_DEVICE_UNKNOWN             0x00000022
#define METHOD_NEITHER                  3
#define FILE_ANY_ACCESS                 0
#define CTL_CODE( DeviceType, Function, Method, Access ) (  \
((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method) \
)

#define htons(a) (USHORT)((((a) & 0xff) << 8) + (((a) & 0xff00) >> 8))

#include "..\shared.h"

#define HOOK_DEBUG 1             /* Trace our hooking of NdisSend()?  */

#define BILLGPS_Major 1
#define BILLGPS_Minor 31
#define BILLGPS_DeviceID UNDEFINED_DEVICE_ID

#define BILLGPS_NDIS_MAJOR_VERSION	0x03
#define BILLGPS_NDIS_MINOR_VERSION	0x00

#define TRANSPORT_NAME "BILLGPS"

#define MAX_REQUESTS 4
#define MAX_TRANSMITS 2

typedef struct _BgpsReserved   { // must be <= 16 bytes for receive packet
  LIST_ENTRY listElement;
  PIRP irpP;
  PMDL mdlP;
}
BgpsReserved, *PBgpsReserved;

#define RESERVED(_p) ((PBgpsReserved)((_p)->ProtocolReserved))

typedef struct _BgpsReqNtry 
{
  LIST_ENTRY listElement;
  PIRP irpP;                    // associated IRP, if any
  NDIS_REQUEST request;
} 
BgpsReqNtry, *PBgpsReqNtry;

/*
  Device extension
*/

/*
  Port device extension.  The symbolicLink buffer points
  into the deviceName buffer, so it mustn't be freed.
*/

typedef struct _DeviceExtension {
  PDEVICE_OBJECT deviceObjectP;
  NDIS_STRING macName;
  NDIS_STRING deviceName;
  NDIS_STRING symbolicLink;
} DeviceExtension, *PDeviceExtension;

enum { TokenRingRoutingInfoIndicator = 0x80 }; // applied to TR source address

typedef struct _TokenRingHeader { // for our output: broadcast w/routing info
  UCHAR accessControl;
  UCHAR frameControl;
  UCHAR destination[MacAddressSize];
  UCHAR source[MacAddressSize];
  UCHAR routingInfo[2];
  LLCSNAP llcSnap;
} TokenRingHeader;

typedef union _MacHeaderPayload {
  TokenRingHeader tr;
  RFC894Header eth;
} MacHeaderPayload;             // room for Ethernet or Token Ring

typedef struct _TxMacHeaderNtry {
  LIST_ENTRY listElement;       // must be first!
  MacHeaderPayload payload;     // room for Ethernet or Token Ring
} TxMacHeaderNtry, *PTxMacHeaderNtry;

typedef struct _TxIpUdpHeader {
  IPHeader ip;
  UDPHeader udp;
} TxIpUdpHeader;

typedef struct _TxIpUdpHeaderNtry {
  LIST_ENTRY listElement;       // must be first!
  TxIpUdpHeader payload;
} TxIpUdpHeaderNtry, *PTxIpUdpHeaderNtry;

typedef struct _OpenInstance 
{
  PDEVICE_OBJECT deviceObjectP;
  PIRP openOrCloseIrpP;         // convenience for xxCompleteHandler
  PIRP cleanupIrpP;             // IRP for MJ_CLEANUP dispatch routine
  NDIS_HANDLE adapterHandle;    // returned from NdisOpenAdapter
  NDIS_SPIN_LOCK requestSpinLock;
  LIST_ENTRY requestList;
  BgpsReqNtry requests[MAX_REQUESTS];
  NDIS_SPIN_LOCK txMacHeaderSpinLock;
  LIST_ENTRY txMacHeaderList;
  TxMacHeaderNtry txMacHeaders[MAX_TRANSMITS];
  NDIS_SPIN_LOCK txIpUdpHeaderSpinLock;
  LIST_ENTRY txIpUdpHeaderList;
  TxIpUdpHeaderNtry txIpUdpHeaders[MAX_TRANSMITS];
  NDIS_HANDLE packetPool;
#if 0                           // not needed in NT?
  NDIS_HANDLE bufferPool;
#endif

  LIST_ENTRY rcvList;

#if 0                           // These VPACKET things we don't need:


  NDIS_SPIN_LOCK ResetSpinLock;

  INTERNAL_REQUEST Requests[MAX_REQUESTS];

#endif

  UCHAR macAddress[MacAddressSize];
  union {
    NDIS_MEDIUM medium;
    UINT selectedMediumIndex;
  } u1;
  UINT packetFilter;
  int promiscuous;              // are we accepting any old packet?
  /* Now come statuses that we keep around so we can dump 'em: */
  NDIS_STATUS openErrorStatus;
  NDIS_STATUS macAddressReqStatus;
  NDIS_STATUS setFilterReqStatus;
  NDIS_STATUS transferDataStatus;
} 
OpenInstance, *POpenInstance;

#define  NPackets 16            // total # of packets for xmit & receive

NTSTATUS DriverEntry(IN PDRIVER_OBJECT adriverObjectP,
                     IN PUNICODE_STRING aregistryPathP);

/*
  Functions called out in the driver object:
*/

VOID BgpsUnload(IN PDRIVER_OBJECT adriverObjectP);

NTSTATUS BgpsOpen(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP);

NTSTATUS BgpsClose(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP);

NTSTATUS BgpsWrite(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP);

NTSTATUS BgpsRead(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP);

NTSTATUS BgpsIoControl(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP);

NTSTATUS BgpsCleanup(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP);

/*
   NDIS 3.0 PROTOCOL FUNCTION PROTOTYPES
*/

VOID
BgpsOpenAdapterCompleteHandler(IN NDIS_HANDLE abindingContext,
                               IN NDIS_STATUS astatus,
                               IN NDIS_STATUS aopenErrorStatus);

VOID
BgpsCloseAdapterCompleteHandler(IN NDIS_HANDLE abindingContext,
                                IN NDIS_STATUS astatus);

VOID
BgpsSendCompleteHandler(IN NDIS_HANDLE abindingContext,
                        IN PNDIS_PACKET apacket,
                        IN NDIS_STATUS astatus);

VOID
BgpsTransferDataCompleteHandler(IN NDIS_HANDLE abindingContext,
                                IN PNDIS_PACKET apacket,
                                IN NDIS_STATUS astatus,
                                IN UINT abytesTransferred);

VOID
BgpsResetCompleteHandler(IN NDIS_HANDLE abindingContext,
                         IN NDIS_STATUS astatus);

VOID
BgpsRequestCompleteHandler(IN NDIS_HANDLE abindingContext,
                           IN PNDIS_REQUEST andisRequest,
                           IN NDIS_STATUS astatus);

NDIS_STATUS
BgpsReceiveHandler(IN NDIS_HANDLE abindingContext,
                   IN NDIS_HANDLE amacReceiveContext,
                   IN PVOID aheaderBuffer,
                   IN UINT aheaderBufferSize,
                   IN PVOID alookAheadBuffer,
                   IN UINT alookaheadBufferSize,
                   IN UINT apacketSize);

VOID
BgpsReceiveCompleteHandler(IN NDIS_HANDLE NdisBindingContext);

VOID
BgpsStatusHandler(IN NDIS_HANDLE abindingContext,
                  IN NDIS_STATUS ageneralStatus,
                  IN PVOID astatusBuffer,
                  IN UINT astatusBufferSize);

VOID
BgpsStatusCompleteHandler(IN NDIS_HANDLE abindingContext);

/*
  Non-NDIS Prototypes
*/

NDIS_STATUS
BgpsAllocatePacketBuffer(POpenInstance pOpen,
                         PNDIS_PACKET *pPacket, // 17.2.97
                         PIO_STACK_LOCATION aPISL);

void BgpsDisconnect(PIO_STACK_LOCATION aPISL);

void BgpsCancelReadIrp(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP);

/*
   PROTOCOL CONTEXT

   We only have one of these structures for the Protocol VxD.
   When updating this structure, remember that there's a reason for the
   way its members are ordered: the needs of the IOCTL to dump the driver's
   guts.  Make sure the global stuff comes first, followed by the adapter
   contexts, and lastly (what doesn't need to be dumped) the packet buffer.
*/

typedef struct _ProtocolBlock {
  NDIS_HANDLE ndisProtocolHandle;     // returned from NdisRegisterProtocol
} ProtocolBlock, *PProtocolBlock;

extern ProtocolBlock billgpsContext;

#endif // __BILLGPS_H__


